import base64
import logging
import os
import uuid
import webbrowser

import requests as r
from spotckup.utils import do_request_validate_response, path_or_create


def auth(client_id: str, client_secret: str, dir_path: str, redirect_uri: str, debug: bool, verbose: bool):
    logging.basicConfig(format='[%(levelname)s] %(message)s')
    log: logging.Logger = logging.getLogger('')
    if debug: log.setLevel('DEBUG')

    scopes: str = 'playlist-read-collaborative ' \
                  'playlist-modify-public ' \
                  'playlist-read-private ' \
                  'playlist-modify-private ' \
                  'ugc-image-upload ' \
                  'user-library-modify ' \
                  'user-library-read '

    log.info("Client id: " + client_id + "\tClient secret: " + client_secret)
    random_state = uuid.uuid4()
    log.debug("Generated state: " + random_state.__str__())

    res: r.Response = do_request_validate_response('GET', 'https://accounts.spotify.com/authorize', verbose=verbose,
                                                   params={
                                                       'client_id': client_id,
                                                       'response_type': 'code',
                                                       'redirect_uri': redirect_uri,
                                                       'state': random_state,
                                                       'scope': scopes
                                                   })

    print('Opening browser... Log in spotify and paste here the url the browser redirects you to.')
    webbrowser.open(res.url, new=2)
    res_url = str(input("Url: - something like redirect_uri?code=AQB-w...&state=3ae...\n"))
    tokens = {k: v for k, v in [qp.split("=", 2) for qp in res_url.split("?", 1)[1].split("&")]}

    if 'error' in tokens: raise Exception("The browser did not authorize the request")

    log.debug('Received state: ' + tokens['state'])
    log.info('Authorization code: ' + tokens['code'])
    if tokens['state'] != random_state.__str__(): raise Exception("Invalid state")

    res: r.Response = do_request_validate_response('POST', 'https://accounts.spotify.com/api/token', verbose=verbose,
                                                   data={
                                                       'grant_type': "authorization_code",
                                                       'code': tokens['code'],
                                                       'redirect_uri': redirect_uri
                                                   }, headers={
            "Authorization": "Basic " + str(
                base64.b64encode(bytes(client_id + ':' + client_secret, 'utf-8')).decode('utf-8'))
        })

    res_json: {} = res.json()
    log.info('Access token: ' + res_json['access_token'])
    log.info('Scopes: ' + res_json['scope'])
    log.info('Expires in: ' + str(res_json['expires_in']))
    log.info('Refresh token: ' + res_json['refresh_token'])

    path: str = path_or_create(dir_path)
    with open(f"{path}/access_token", "w+") as f:
        f.write(res_json['access_token'])

    with open(f"{path}/refresh_token", "w+") as f:
        f.write(res_json['refresh_token'])

    print("Done. Tokens saved in files access_token and refresh_token.\n" +
          "You may now want to run $ spotckup backup or set the cron job for automatic backup." +
          "Read the README file for more infos.")
    exit(0)



import json
import logging
import os
from typing import List, Dict

from spotckup.decorators import timer
from spotckup.utils import save_image_from_url, do_request_validate_response, path_or_create


@timer
def get_list_from_paginated_response(url: str, token: str, verbose: bool = False) -> List[Dict]:
    res: List = []
    log: logging.Logger = logging.getLogger('')
    while url is not None:
        next_res = do_request_validate_response('GET', url, verbose=verbose, headers={
            "Authorization": "Bearer " + token
        }).json()
        log.debug(next_res['next'])
        res = res + next_res['items']
        url = next_res['next']
    return res


def backup_playlist(authorization_token, dir_path, debug, verbose):
    logging.basicConfig(format='[%(levelname)s] %(message)s')
    log: logging.Logger = logging.getLogger('')
    if debug: log.setLevel('DEBUG')

    path: str = path_or_create(dir_path)

    token: str = authorization_token
    if token is None:
        with open(f'{path}/access_token', 'r') as f:
            token = f.read()

    user_id: str = do_request_validate_response('GET', 'https://api.spotify.com/v1/me',
                                                verbose=verbose,
                                                headers={
                                                    "Authorization": "Bearer " + token
                                                }).json()['id']

    playlists: [] = get_list_from_paginated_response(f'https://api.spotify.com/v1/users/{user_id}/playlists', token,
                                                     verbose=verbose)
    # playlists: {} = do_request_validate_playlistsponse('GET', 'https://api.spotify.com/v1/users/{}/playlists'.format(user_id),
    #                                        verbose=verbose,
    #                                        params={
    #                                            'limit': 50,
    #                                            'offset': 0},
    #                                        headers={"Authorization": "Bearer " + token
    #                                                 }).json()
    print('Fetched {} playlists.'.format(str(len(playlists))))

    os.makedirs(os.path.dirname(f"{path}/img/"), exist_ok=True)
    for playlist_meta in playlists:
        if playlist_meta['images']:
            if not os.path.exists(f'{path}/img/{playlist_meta["id"]}.jpg'):
                save_image_from_url(playlist_meta['images'][0]['url'], playlist_meta['id'], f'{path}/img')

    with open(f'{path}/playlists-metadata.json', 'w+') as f:
        json.dump(playlists, f, indent=4)

    log.info('Succesfully wrote {} playlists metadata in playlists-metadata.json'.format(str(len(playlists))))

    with open(f'{path}/playlist.json', 'w') as f:
        json.dump({
            (playlist['id'] + '#' + playlist['snapshot_id']): get_list_from_paginated_response(
                'https://api.spotify.com/v1/playlists/{}/tracks?fields=next,items(is_local,track(name,uri,album(name),artists(name),artist(name)))'
                    .format(playlist['id']), token, verbose=verbose)
            for playlist in playlists
        }, f, indent=4)

    print('The playlists backup has completed succesfully.')



import argparse
import json
import logging

import requests as r

from spotckup.utils import do_request_validate_response, path_or_create


def restore_library(authorization_token: str, dir_path: str, debug: bool, verbose: bool):
    logging.basicConfig(format='[%(levelname)s] %(message)s')
    log: logging.Logger = logging.getLogger('')
    if debug: log.setLevel('DEBUG')

    path = path_or_create(dir_path)

    token: str = authorization_token
    if token is None:
        with open(f'{path}/access_token', 'r') as f:
            token = f.read()

    with open(f'{path}/library.json', 'r') as f:
        library = json.load(f)
        print(f"Read from local backup {len(library)} albums.")

    for i in range(0, len(library), 50):
        chunk = [el['id'] for el in library[i:i + 50]]
        log.debug(chunk)
        do_request_validate_response('PUT',
                                     f'https://api.spotify.com/v1/me/albums',
                                     verbose=verbose,
                                     data=json.dumps({'ids': chunk}),
                                     headers={
                                         "Authorization": "Bearer " + token,
                                         "Content-Type": "application/json"
                                     })

    print('The albums were imported in the spotify library succesfully.')



import logging
import os

import requests as r
from spotckup.decorators import timer


@timer
def save_image_from_url(url: str, img_name: str, path: str) -> None:
    img_res: r.Response = do_request_validate_response('GET', url, stream=True)
    img_res.raw.decode_content = True
    img = img_res.content  # Binary content!
    with open("{}/{}.jpg".format(path, img_name), 'wb') as img_file:
        img_file.write(img)


@timer
def do_request_validate_response(method: str, url: str, verbose: bool = False, **kwargs) -> r.Response:
    log = logging.getLogger('')
    log.debug(f"Requesting: {method} {url}")
    log.debug(str(verbose))
    if verbose:
        log.debug([str(k) + ':\t' + v + '\n' for k, v in kwargs.items()])
    res = r.request(method, url, **kwargs)
    log.debug(f'Response status code: {res.status_code}')
    if verbose:
        log.debug(res.content)
    if res.status_code not in range(200, 300):
        if res.status_code == 429:
            raise Exception("Too many calls. Retry after {}".format(res.headers.get('Retry-After')))
        raise Exception(
            "StatusCode: {}\nError: {}".format(res.status_code, res.text))
    return res


def path_or_create(dir_path: str) -> str:
    path = os.path.join(os.path.expanduser('~'), 'Documents/spotckup/data')
    if dir_path is not '' and dir_path is not None:
        path = dir_path
    logging.getLogger('').info(path)

    if os.path.exists(path):
        if not os.path.isdir(path):
            raise Exception(f"The specified path '{path}' is not a directory")
    else:
        create_flag: bool = True if str(
            input(f"The specified path '{path}' does not exist. Do you want to create it? (y/[n])")) == 'y' else False
        if create_flag:
            os.makedirs(path, exist_ok=True)
        else:
            print("Did not create the directory. Exiting application...")
            exit(1)
    return path



import base64
import json
import logging
import os

import requests as r

from spotckup.utils import do_request_validate_response, path_or_create


def restore_playlist(authorization_token: str, dir_path: str, debug: bool, verbose: bool):
    logging.basicConfig(format='[%(levelname)s] %(message)s')
    log: logging.Logger = logging.getLogger('')
    if debug: log.setLevel('DEBUG')

    path: str = path_or_create(dir_path)

    token: str = authorization_token
    if token is None:
        with open(f'{path}/access_token', 'r') as f:
            token = f.read()

    user_id: str = do_request_validate_response('GET', 'https://api.spotify.com/v1/me', verbose=verbose, headers={
        "Authorization": "Bearer " + token
    }).json()['id']

    with open(f'{path}/playlists-metadata.json', 'r') as f:
        playlists_metadata = json.load(f)
        print(f"Read from local backup {len(playlists_metadata)} playlists.")

    link_metadata: {} = {}
    imgs_urls: {} = {}

    for playlist_meta in playlists_metadata:
        if len(playlist_meta) == 0: break
        res: r.Response = do_request_validate_response('POST',
                                                       'https://api.spotify.com/v1/users/{}/playlists'.format(user_id),
                                                       verbose=verbose,
                                                       data=json.dumps({
                                                           'name': playlist_meta['name'],
                                                           'public': str(playlist_meta['public']),
                                                           'collaborative': str(playlist_meta['collaborative']),
                                                           'description': playlist_meta['description']
                                                       }),
                                                       headers={
                                                           "Authorization": "Bearer " + token,
                                                           "Content-type": "application/json"
                                                       })
        res_json = res.json()
        link_metadata[playlist_meta['id']] = res_json['id']
        if len(playlist_meta['images']) != 0:
            log.debug('Playlist cover: ' + playlist_meta['images'][0]['url'])
            imgs_urls[playlist_meta['id']] = playlist_meta['images'][0]['url']
        log.debug(f'Succesfuly created playlist {playlist_meta["name"]}.\nStatus code: {res.status_code}')
        log.debug(f"New id: {res.json()['id']}")
        del res

    with open(f'{path}/playlist.json', 'r') as f:
        playlists = json.load(f)
        for playlist in playlists:
            old_id = playlist.split('#')[0]
            id = link_metadata[old_id]
            img = None

            if os.path.exists(f'{path}/img/{old_id}.jpg') and os.path.isfile(f'{path}/img/{old_id}.jpg'):
                with open(f'{path}/img/{old_id}.jpg', 'rb') as img_file:
                    img = img_file.read()

            playlist_no_local = list(filter(lambda not_local: not_local['track'] is not None,
                                            filter(lambda track: not track['is_local'],
                                                   playlists[playlist])))
            for i in range(0, len(playlist_no_local), 100):
                chunk = [track['track']['uri'] for track in playlist_no_local[i:i + 100]]
                do_request_validate_response('POST',
                                             f'https://api.spotify.com/v1/playlists/{id}/tracks',
                                             verbose=verbose,
                                             data=json.dumps({
                                                 "uris": chunk
                                             }),
                                             headers={
                                                 "Authorization": "Bearer " + token,
                                                 "Content-Type": "application/json"
                                             })

            if img is not None:
                do_request_validate_response('PUT',
                                             f'https://api.spotify.com/v1/playlists/{id}/images',
                                             verbose=verbose,
                                             data=base64.b64encode(img).decode('utf-8'),
                                             headers={
                                                 "Authorization": "Bearer " + token,
                                                 "Content-Type": "image/jpeg"
                                             })

    print('The playlists were imported on spotify succesfully.')



import click
import os
import shutil

from spotckup import authorize, refresh as ref, library, playlists, restore_playlists, restore_library

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version='0.1.0')
def spotckup():
    pass


@spotckup.command()
def clear():
    if str(input(
            "Attention! You are going to delete all local files, including access/refresh tokens, cover images and "
            "playlists/library from default directory ~/Documents/spotckup.\n\n"
            "Are you sure? (y/[n]):\t")) == 'y':
        shutil.rmtree(os.path.join(os.path.expanduser('~'), 'Documents/spotckup'), ignore_errors=True)


@spotckup.command()
@click.option('--client-id', default=os.getenv('SPOTIFY_CLIENT_ID'))
@click.option('--client-secret', default=os.getenv('SPOTIFY_CLIENT_SECRET'))
@click.option('--redirect-uri', default='http://localhost', help='The redirect uri defined in the spotify web app.'
                                                                 + ' If the provided uri does not match the spotify one '
                                                                 + 'the authorization will fail.')
@click.option('-p', '--path', default='',
              help='The path to the directory where you want to store tokens and backup data')
@click.option('--debug', '-d', default=False, is_flag=True, help='Enable debug output.')
@click.option('--verbose', default=False, is_flag=True, help='Makes debug log more verbose.')
def auth(client_id, client_secret, path, redirect_uri, debug, verbose):
    validate_id_and_secret(client_id, client_secret)
    authorize.auth(client_id, client_secret, path, redirect_uri, debug, verbose)


@spotckup.command()
@click.option('--token', '-t', default=None,
              help='The access token, if this option is not present it will try to read it '
                   + 'from data/access_token')
@click.option('-p', '--path', default='',
              help='The path to the directory where you want to store tokens and backup data')
@click.option('--debug', '-d', default=False, is_flag=True, help='Enable debug output.')
@click.option('--verbose', default=False, is_flag=True, help='Makes debug logs more verbose.')
@click.option('--only-library', default=False, is_flag=True, help='Set this flag to backup only the library')
@click.option('--only-playlist', default=False, is_flag=True, help='Set this flag to backup only the playlists')
def backup(token, path, debug, verbose, only_library, only_playlist):
    if not only_playlist: library.backup_library(token, path, debug, verbose)
    if not only_library: playlists.backup_playlist(token, path, debug, verbose)
    exit(0)


@spotckup.command()
@click.option('--token', '-t', default=None,
              help='The access token, if this option is not present it will try to read it '
                   + 'from data/access_token')
@click.option('-p', '--path', default='',
              help='The path to the directory where you want to store tokens and backup data')
@click.option('--debug', '-d', default=False, is_flag=True, help='Enable debug output.')
@click.option('--verbose', default=False, is_flag=True, help='Makes debug logs more verbose.')
@click.option('--only-library', default=False, is_flag=True, help='Set this flag to restore only the library')
@click.option('--only-playlist', default=False, is_flag=True, help='Set this flag to restore only the playlists')
def restore(token, path, debug, verbose, only_library, only_playlist):
    if not only_playlist: restore_library.restore_library(token, path, debug, verbose)
    if not only_library: restore_playlists.restore_playlist(token, path, debug, verbose)
    exit(0)


@spotckup.command()
@click.option('--client-id', default=os.getenv('SPOTIFY_CLIENT_ID'))
@click.option('--client-secret', default=os.getenv('SPOTIFY_CLIENT_SECRET'))
@click.option('--token', '-t', default=None,
              help='The refresh token, if this option is not present it will try to read it '
                   + 'from data/refresh_token')
@click.option('-p', '--path', default='',
              help='The path to the directory where you want to store tokens and backup data')
@click.option('--debug', '-d', default=False, is_flag=True, help='Enable debug output.')
@click.option('--verbose', default=False, is_flag=True, help='Makes debug log more verbose.')
def refresh(client_id, client_secret, token, path, debug, verbose):
    validate_id_and_secret(client_id, client_secret)
    ref.refresh_token(client_id, client_secret, path, token, debug, verbose)


def validate_id_and_secret(client_id: str, client_secret: str):
    if client_id is None or client_id is '':
        print("Client ID is mandatory. You can either set an environment variable named SPOTIFY_CLIENT_ID "
              + "or pass it with the --client-id flag.")
        exit(1)
    if client_secret is None or client_secret is '':
        print("Client secret is mandatory. You can either set an environment variable named SPOTIFY_CLIENT_SECRET "
              + "or pass it with the --client-secret flag.")
        exit(1)


if __name__ == '__main__':
    spotckup()



import base64
import logging

import requests as r

from spotckup.utils import do_request_validate_response, path_or_create


def refresh_token(client_id: str, client_secret: str, dir_path: str, refresh_token: str, debug: bool, verbose: str):
    logging.basicConfig(format='[%(levelname)s] %(message)s')
    log: logging.Logger = logging.getLogger('')
    if debug: log.setLevel('DEBUG')

    log.info("Client id: " + client_id + "\tClient secret: " + client_secret)

    path: str = path_or_create(dir_path)

    token: str = refresh_token
    if token is None:
        with open(f'{path}/refresh_token', 'r') as f:
            token = f.read()

    res: r.Response = do_request_validate_response('POST', 'https://accounts.spotify.com/api/token', verbose=verbose,
                                                   data={
                                                       'grant_type': "refresh_token",
                                                       'refresh_token': token
                                                   }, headers={
            "Authorization": "Basic " + str(
                base64.b64encode(bytes(client_id + ':' + client_secret, 'utf-8')).decode('utf-8'))
        })

    res_json: {} = res.json()
    log.info('Access token: ' + res_json['access_token'])
    log.info('Scopes: ' + res_json['scope'])
    log.info('Expires in: ' + str(res_json['expires_in']))

    with open(f"{path}/access_token", "w+") as f:
        f.write(res_json['access_token'])

    print("Token refreshed.")
    exit(0)



import json
import logging
from typing import List

from spotckup.utils import do_request_validate_response, path_or_create


def get_tracks_from_albums(url: str, token: str, verbose: bool = False):
    res: List = []
    while url is not None:
        next_res = do_request_validate_response('GET', url, verbose=verbose, headers={
            "Authorization": "Bearer " + token
        }).json()
        res = res + [
            {
                'id': album['album']['id'],
                'name': album['album']['name'],
                'artists': list(map(lambda a: a['name'], album['album']['artists']))
            }
            for album in next_res['items']
        ]
        url = next_res['next']
        log: logging.Logger = logging.getLogger('')
        log.debug(url)
    return res


def backup_library(authorization_token, dir_path: str, debug, verbose):
    logging.basicConfig(format='[%(levelname)s] %(message)s')
    log: logging.Logger = logging.getLogger('')
    if debug: log.setLevel('DEBUG')

    path: str = path_or_create(dir_path)

    token: str = authorization_token
    if token is None:
        with open(f'{path}/access_token', 'r') as f:
            token = f.read()

    print("Fetching albums from library...")
    with open(f'{path}/library.json', 'w') as f:
        json.dump(get_tracks_from_albums('https://api.spotify.com/v1/me/albums?limit=50', token, verbose=verbose)
                  , f, indent=4)

    print('Successful backup of all library.')



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="spotckup",
    version="0.1.0",
    author="mattiaforc",
    author_email="mattiaforc@gmail.com",
    description="A small utility for creating JSON local backups of music and playlists from spotify",
    long_description=long_description,
    url="https://github.com/mattiaforc/spotckup",
    keywords="spotify backup utility json store",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'Click',
        'requests',
        'certifi',
        'idna',
        'urllib3'
    ],
    entry_points={
        'console_scripts': [
            'spotckup=spotckup.cli:spotckup',  # command=package.module:function
        ],
    },

)



import functools
import logging
import time


def timer(func):
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        log = logging.getLogger('')
        start_time = time.perf_counter()
        value = func(*args, **kwargs)
        end_time = time.perf_counter()
        run_time = end_time - start_time
        log.info(f'Finished in {run_time:.16f} secs')
        return value

    return wrapper_timer
